import pyLDAvis.gensim
import pyLDAvis
from gensim import corpora, models


class LDA:
    """
    performs Latent Dirichlet Allocation Topic Modelling
    """

    def __init__(self):
        3 + 4

    @staticmethod
    def preperation(df_processed, no_below=None, no_above=None):
        """Preprocessing for LDA

        Args:
            df_processed ([DataFrame]): [DataFrame stat store tokenized text in
                                        col ["tokens"]]
            no_below ([int]): [if word appears in less than
                              no_below documents they are not considered]
            no_above ([float]): [if e.g. 0.9, no words are taken into
                                account that appear more often than 90%]

        Returns:
            [dictionary]: [returns the dictionary used for LDA]
            [bow_corpus]: [Corpus used for LDA]
        """

        dictionary = corpora.Dictionary(df_processed["tokens"])
        if no_below is None:
            no_below = 0
        if no_above is None:
            no_above = 1
        bow_corpus = [dictionary.doc2bow(doc) for doc
                      in df_processed["tokens"]]
        return dictionary, bow_corpus

    @staticmethod
    def model(corpus, num_topics, id2word, random_state, passes):
        """LDA model

        Args:
        corpus (iterable of list of (int, float), optional):
        [Stream of document vectors or sparse matrix of shape]
        num_topics (int): [pre-defined number of topics]
        id2word ({dict of (int, str): gensim.corpora.dictionary.Dictionary})
                                         –Mapping from word IDs to words. It is
                                        used to determine the vocabulary size,
                                as well as for debugging and topic printing.
        random_state (int): [for recreating exact identical output]
        passes (int): [Number of passes through the corpus during training.]

        Returns:
            [lda_model]: [returns lda_model output]
        """

        lda_model = models.LdaMulticore(
            corpus,
            num_topics=num_topics,
            id2word=id2word,
            random_state=random_state,
            passes=passes,
        )

        return lda_model

    @staticmethod
    def visualize_topics(lda_model, bow_corpus, dictionary):
        """Create pyLDAvis plots for LDA model output

        Args:
            lda_model ([type]): Output of lda_model
            bow_corpus ([list of (int, int)]): BoW representation of document.
            dictionary (gensim.corpora.dictionary.Dictionary): [Dict used for
                                                               creating Corpus]
        """

        visualization = pyLDAvis.gensim.prepare(
            lda_model, bow_corpus, dictionary, sort_topics=False
        )

        pyLDAvis.show(visualization)



import pandas as pd


papers = pd.read_csv(r"tests\dentistry_teeth.txt")
papers.head()
papers = papers.drop_duplicates(subset=["text"])
mistake = papers["text"].iloc[77]
papers = papers[papers["text"] != mistake]
if __name__ == "__main__":
    print(papers)



import re
import webbot
import time
from bs4 import BeautifulSoup
import numpy as np
import pandas as pd
import requests
import json


class AbstractScraper:
    """
    Abstractscraper - scrapes through IEEE Xplore and, depending on the given
    search query, scrapes the abstracts of scientific papers
    """

    def __init__(self):
        self.web = webbot.Browser()
        self.wait = 5
        self.first = "https://ieeexplore.ieee.org/search/searchresult.jsp?action=search&newsearch=true&matchBoolean=true&"
        self.second = "&highlight=true&returnFacets=ALL&returnType=SEARCH&matchPubs=true&rowsPerPage=100&pageNumber=1"

    def _open(self, url=None, keywords=None, operator="OR", in_data="author",
              pages=2):
        """
        defines the users search query

        :param url: when the user specifies an own search query on IEEEXplore
        :type url: string
        :param keywords: keywords that are searched for
        :type keywords: list
        :param operator: "and" / "or" operator between keywords
        :type operator: string
        :param pages: number of pages that is iterated over
        :type pages: int
        :param in_data: "author" or "all_meta" whether to search in author
                        keywords or all metadata
        :type in_data: string
        :return:
        """
        operator = ")%20" + operator.upper() + "%20("
        if in_data == "all_meta":
            in_data = "%22All%20Metadata%22:"
        elif in_data == "author":
            in_data = "%22Author%20Keywords%22:"
        if url:
            self.url = url
            self.web.go_to(self.url)

        else:
            if len(keywords) > 3:
                print("You can only specify 3 Keywords/phrases")
            elif len(keywords) == 1:
                self.middle = "queryText=" + in_data + keywords[0]
            elif len(keywords) == 2:
                self.middle = (
                    "queryText=(("
                    + in_data
                    + keywords[0]
                    + operator
                    + in_data
                    + keywords[1]
                    + "))"
                )
            else:
                self.middle = (
                    "queryText=(("
                    + in_data
                    + keywords[0]
                    + operator
                    + in_data
                    + keywords[1]
                    + operator
                    + in_data
                    + keywords[2]
                    + "))"
                )
            self.url = self.first + self.middle + self.second
            self.web.go_to(self.url)
            time.sleep(self.wait)

        """goes through the number of defined pages on IEEE Xplore"""
        # go to web page
        self.web.click("Accept")
        self.html_page = []

        for i in range(1, pages + 1):
            self.web.click(str(i))
            time.sleep(self.wait)
            self.html_page.append(self.web.get_page_source())
            time.sleep(self.wait)
        print("The algorithm is iterating through", len(self.html_page),
              "pages")
        self.web.quit()

    def _find_links(self):
        """goes through every paper on every page and collects all links to the
         subpages of the papers"""
        document_links = []
        self.data = []
        for j in range(len(self.html_page)):
            for link in BeautifulSoup(
                self.html_page[j], features="html.parser"
            ).findAll("a", attrs={"href": re.compile("^/document")}):
                document_links.append(link.get("href"))

        # remove unnecessary results of the href search
        matching = [s for s in document_links if "citation" in s]
        x = [i for i in document_links if i not in matching]
        self.data = x
        self.data = np.array(self.data)
        # remove duplicates that are in there due to multiple occurrence in the
        self.data = np.unique(self.data)
        print("Total number of abstracts that will be scraped:",
              len(self.data))

    def _find_abstracts(self):
        """Opens all links for the webpages for each paper and scrapes the
        paper's abstract"""
        # initialize empty lists to fill
        self.abstracts = []
        self.title = []

        # loop through number of  "link"
        for i in range(len(self.data)):
            # only "try" because sometimes the javascript is corrupted
            try:
                data = json.loads(
                    re.search(
                        r"\.metadata=(.*?);",
                        requests.get("https://ieeexplore.ieee.org" +
                                     self.data[i]).text,
                    ).group(1)
                )
                # only get title and abstracts -> we could also go for
                # author etc.
                title = data["title"]
                data = data["abstract"]

                self.title.append(title)
                self.abstracts.append(data)
            except:
                pass

    def get_abstracts(
        self, url=None, keywords=None, operator="OR", pages=2, in_data="author"
    ):
        """
        :param url: when the user specifies an own search query on IEEEXplore
        :type url: string
        :param keywords: keywords that are searched for
        :type keywords: list
        :param operator: "and" / "or" operator between keywords
        :type operator: string
        :param pages: number of pages that is iterated over
        :type pages: int
        :param in_data: "author" or "all_meta" whether to search in author
                keywords or all metadata
        :type in_data: string
        :return: pd.DataFrame
        """
        self._open(
            url=url, keywords=keywords, operator=operator, pages=pages,
            in_data=in_data
        )
        self._find_links()
        self._find_abstracts()

        data = pd.DataFrame({"text": self.abstracts, "titles": self.title})
        data = data.drop_duplicates()
        return data


if __name__ == "__main__":
    # Execute the above code
    AS = AbstractScraper()
    data = AS.get_abstracts(
        keywords=["dentistry", "teeth", "tooth"],
        in_data="all_meta",
        pages=12,
        operator="or",
    )
    print(data["text"])



from re import compile as re_compile
from nltk.tokenize import RegexpTokenizer
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer


class Preprocessor:
    def __init__(self):
        4 + 5

    def _text_prepare(text, language="english"):
        """prepares text for lda and o-svm.
        Removes stopwords, symbols, double spaces and numbers

        Args:
            text (numpy array): [column out of dataframe with documents]
            language (string): [language for stopword removel,
                                default: english]

        Returns:
            [numpy array]: [prepared documents]
        """

        REPLACE_BY_SPACE_RE = re_compile("[/(){}\[\]\|@,;]")
        BAD_SYMBOLS_RE = re_compile("[^0-9a-z #+_]")
        NUMBERS = re_compile(r"\d+")
        STOPWORDS = set(stopwords.words(language))

        text = text.lower()
        # replace REPLACE_BY_SPACE_RE symbols by space in text
        text = REPLACE_BY_SPACE_RE.sub("", text)
        text = BAD_SYMBOLS_RE.sub("", text)
        text = NUMBERS.sub("", text)
        # delete symbols which are in BAD_SYMBOLS_RE from text
        words = text.split()
        i = 0
        while i < len(words):
            if words[i] in STOPWORDS:
                words.pop(i)
            else:
                i += 1
        text = " ".join(map(str, words))  # delete stopwords from text

        return text

    def _lemmatize_text(text):
        """helper function

        Args:
            text (numpy array): [dataframe column where documents are stored]

        Returns:
            [numpy array]: [lemmatized documents]
        """
        lemmatizer = WordNetLemmatizer()
        return " ".join([lemmatizer.lemmatize(w, "v") for w in text])

    def basic_preprocessing(df, column):
        """calls helper functions from above and preprocesses documents
        removes all stopwords and unnecessary terms, e.g. symbols, numbers,
        double spaces
        and tokenizes documents

        Args:
            df (DataFrame): [dataframe where documents are stored in one
                             column]
            columns (String): [column name where documents are stored]

        Returns:
            [DataFrame]: [dataframe with preprocessed documents. 2 new columns
                          are appended. The completely preprocessed
            documents are stored in column ['tokens'], the only lemmatized
                                    documents are stored in columns ['lemma']]
        """
        df_temp = df.copy(deep=True)

        df_temp.loc[:, column] = [
            Preprocessor._text_prepare(x, "english")
            for x in df_temp[column].values
        ]
        df_temp[column] = [
            item for item in df_temp[column] if not item.isdigit()
        ]

        tokenizer = RegexpTokenizer(r"\w+")

        df_temp["tokens"] = df_temp[column].apply(tokenizer.tokenize)
        df_temp["lemma"] = df_temp["tokens"].apply(
            Preprocessor._lemmatize_text
            )
        df_temp["tokens"] = df_temp["lemma"].apply(tokenizer.tokenize)

        return df_temp



from numpy import round as np_round
from numpy import arange as np_arange
from AuDoLab.subclasses import abstractscraper
from AuDoLab.subclasses import lda
from AuDoLab.subclasses import one_class_svm
from AuDoLab.subclasses import preprocessing
from AuDoLab.subclasses import tf_idf


class AuDoLab:
    def __init__(self):
        4 + 4

    def scrape_abstracts(self, url, file_name):
        """Function to scrap abstracts of scientific papers from the givin url.
        We used https://ieeexplore.ieee.org/search/advanced to generate a
        list like https://ieeexplore.ieee.org/search/searchresult.jsp?action=se
        arch&newsearch=true&matchBoolean=true&queryText=(%22Author%20Keywords%22
        :cotton)&highlight=true&returnFacets=ALL&returnType=SEARCH&matchPubs=Tru
        e&rowsPerPage=100&pageNumber=1
        with the search results.
        The abstracts of the papers listet on that list of search results will
        be stored in a .txt file with the givin file name.

        Arguments:
        - url ( string)
        - file_name (string)
        """

        ks = abstractscraper.AbstractScraper(url)

        ks.open()
        ks.find_links()
        self.abstracts = ks.get_abstracts()
        file_name = file_name + ".txt"
        self.abstracts.to_csv(file_name, header=True, index=False)
        return self.abstracts

    def preprocessing(self, data, column):
        """ The data will be lemmatized, tokenized and the stopwords will be
        deleted.

        Arguments:
        - data (<class 'pandas.core.frame.DataFrame'>)
        """
        self.data_processed = preprocessing.Preprocessor.basic_preprocessing(
            data, column
        )
        return self.data_processed

    # def tf_idf_features(self, data, papers, features=8000):
    #     """The function tf_idf_features(...) calculates the tfidf scores, but
    #     return only the <features> amount of words with the highest tfidf
    #     scores.
    #
    #     Arguments:
    #     - data (<class 'pandas.core.frame.DataFrame'>)
    #     - papers (<class 'pandas.core.frame.DataFrame'>)
    #     """
    #     tfidf = tf_idf.Tf_idf()
    #     self.data_tfidf_features, self.papers_tfidf_features =
    #     tfidf.tfidf_features(
    #         data, papers, features=features)
    #     return self.data_tfidf_features, self.papers_tfidf_features

    def tf_idf(self, data, papers, data_column, papers_column, features=None):
        """ The function tf_idf(...) calculates the tfidf scores.

        Arguments:
        - data (<class 'pandas.core.frame.DataFrame'>)
        - papers (<class 'pandas.core.frame.DataFrame'>)
        """
        tfidf = tf_idf.Tf_idf()
        self.data_tfidf, self.papers_tfidf = tfidf.tfidf(
            data, papers, data_column, papers_column, features
        )
        return self.data_tfidf, self.papers_tfidf

    def one_class_svm(
        self,
        training,
        predicting,
        nus=np_round(np_arange(0.001, 0.5, 0.001), 5),
        quality_train=0.85,
        min_pred=0.01,
        max_pred=0.1,
    ):
        """
        This is a one class classifier, that uses the training data (usually the
        papers we scraped earlier) to classify the predicting data. The
        returned df might contain multiple classifier for different parameters
        nu(s).

        Arguments:
        - training (<class 'scipy.sparse.csr.csr_matrix'>)
        - predicting (<class 'scipy.sparse.csr.csr_matrix'>)
        - nus (np.array)
        - quality_train (float)
        - min_pred (float)
        - max_pred (float)
        """
        one_Class_SVM = one_class_svm.One_Class_SVM()
        self.df = one_Class_SVM.classification(
            training=training,
            predicting=predicting,
            nus=nus,
            quality_train=quality_train,
            min_pred=min_pred,
            max_pred=max_pred,
        )
        return self.df

    def choose_classifier(self, df, classifier, i):
        """As mentioned in the description of the function one_class_svm(...)
        multiple classifiers might be return. Now you choose one with which
        you want to continue. The df will then only continue the entries that
        were positivly classified.

        Arguments:
        - df (<class 'pandas.core.frame.DataFrame'>)
        - classifier (<class 'pandas.core.frame.DataFrame'>)
        - i (int)
        """
        return one_class_svm.One_Class_SVM.choose_classifier(df, classifier, i)

    def lda_modeling(
        self,
        data,
        no_below=None,
        no_above=None,
        num_topics=5,
        random_state=101,
        passes=20,
    ):
        """The function performs lda modelling as described in this
        https://www.jmlr.org/papers/volume3/blei03a/blei03a.pdf paper.

        Arguments:
        - data (<class 'pandas.core.frame.DataFrame'>)
        - no_below (int)
        - no_above (int)
        - num_topics (int)
        - random_state (int)
        - passes (int)
        """

        self.dictionary, self.bow_corpus = lda.LDA.preperation(data, no_below,
                                                               no_above)
        self.lda = lda.LDA()
        self.lda_model = self.lda.model(
            self.bow_corpus,
            num_topics=num_topics,
            id2word=self.dictionary,
            random_state=random_state,
            passes=passes,
        )
        return self.lda_model

    def lda_visualize_topics(self):
        """The lda model calculated with the function lda_modeling is visualized
        in an html frame and opened in the standard browser.
        """
        lda.LDA.visualize_topics(self.lda_model, self.bow_corpus,
                                 self.dictionary)



import re
import webbot
import time
from bs4 import BeautifulSoup
import numpy as np
import pandas as pd


sub = " Abstract:\n"


class AbstractScraper:
    """
    Abstractscraper - scrapes through IEEE Xplore and, depending on the given
    search query, scrapes the abstracts of scientific papers
    """

    def __init__(self, url):
        # define url - is already the result of search for the keyword
        # "Artificial Intelligence"
        self.url = url
        self.web = webbot.Browser()
        # define a delay and wait function in order to prevent empty results
        # when internet connection is slow
        self.delay = 3
        self.wait = 6

    def open(self):
        """goes through the number of defined pages on IEEE Xplore"""
        # go to web page
        self.web.go_to(self.url)

        time.sleep(self.delay)
        self.html_page = []
        self.html_page.append(self.web.get_page_source())

        time.sleep(self.wait)

        for i in range(1, 3):
            time.sleep(self.wait)
            self.web.click(str(i))
            time.sleep(self.wait)
            self.html_page.append(self.web.get_page_source())

        print("The algorithm is iterating through", len(self.html_page),
              "pages")
        return self.html_page

    def find_links(self):
        """goes through every paper on every page and collects all links to the
         subpages of the papers"""
        document_links = []
        self.data = []
        for j in range(len(self.html_page)):
            BeautifulSoup(self.html_page[j], features="html.parser")
            time.sleep(self.wait)

            # get the hyperlinks for all the documents and temporarily save
            # them
            for link in BeautifulSoup(
                self.html_page[j], features="html.parser"
            ).findAll("a", attrs={"href": re.compile("^/document")}):
                document_links.append(link.get("href"))
            time.sleep(self.wait)

        # remove unnecessary results of the href search
        matching = [s for s in document_links if "citation" in s]
        x = [i for i in document_links if i not in matching]
        self.data = x
        time.sleep(self.wait)

        self.data = np.array(self.data)
        # remove duplicates that are in there due to multiple occurrence in the
        # href
        self.data = np.unique(self.data)
        print("Total number of abstracts that will be scraped:",
              len(self.data))
        return self.data

    def get_abstracts(self):
        """Opens all links for the webpages for each paper and scrapes
         the paper's abstract"""

        self.abstracts = []

        # go through every search result and do the following: open the
        # keywords section,
        # extract the keywords (+ unnecessary stuff) ,append the keywords to
        # self.keys
        for i in range(len(self.data)):

            self.web.go_to("https://ieeexplore.ieee.org" + self.data[i])
            time.sleep(self.delay)

            html_page = self.web.get_page_source()
            soup = BeautifulSoup(html_page, features="html.parser")
            time.sleep(self.delay)

            texts = soup.find_all("div", {"class": "u-mb-1"})
            for t in texts:
                text = t.text.strip()
                text = text.replace("Abstract:\n", " ")
                df_dict = {"text": text}
            self.abstracts.append(df_dict)

        self.abstracts = pd.DataFrame(self.abstracts)
        self.paper_cleaning()
        return self.abstracts

    def paper_cleaning(self):
        self.abstracts = self.abstracts.drop_duplicates(subset=["text"])
        mistake = self.abstracts["text"].iloc[77]
        self.abstracts = self.abstracts[self.abstracts["text"] != mistake]


if __name__ == "__main__":

    # Execute the above code
    ks = AbstractScraper(
        "https://ieeexplore.ieee.org/search/searchresult.jsp?action=search&newsearch=true&matchBoolean=true&queryText=(%22Author%20Keywords%22:cotton)&highlight=true&returnFacets=ALL&returnType=SEARCH&matchPubs=true&rowsPerPage=100&pageNumber=1"
    )

    html_code = ks.open()
    links = ks.find_links()
    abstracts = ks.get_abstracts()

    abstracts.to_csv("cotton.txt", header=True, index=False)



from AuDoLab.subclasses.preprocessing import Preprocessor
from sklearn.feature_extraction.text import TfidfVectorizer


class Tf_idf:
    """
    Beschreibung
    """

    def __init__(self):
        4 + 5

    @staticmethod
    def tfidf(
        data,
        papers,
        data_column="tokens",
        papers_column="tokens",
        features=None,
        ngrams=2,
    ):
        """[creates tf-idf objects for one-class SVM classification. The tf-idf
        scores are calculated over a joint corpus,
        however, the target data and the out-of-domain training data are stored
        in seperate, as the one-class SVM is only trained on
        the tf-idf scores of the out-of-domain training data]

        Args:
            data (DataFrame): [preprocessed target documents]
            papers (DataFrame): [preprocessed out-of-domain training data]
            data_colum (String): [name of columnin target dataframe where
            lemmatized documents are stored]. Defaults to 'lemma'
            papers_colum (String): [name of column in out-of-domain training
            dataframe where lemmatized documents are stored]. Defaults to
            'lemma' ngrams (int, optional): [whether ngram are formed].
            Defaults to 2.
            features (int, optional): [number of max features].
            Defaults to 8000.

        Returns:
            [data and papers]: [tfidf object data for target data and
             ou-of-domain training data]
        """

        Preprocessor()
        df_temp = data.copy(deep=True)
        df_temp = Preprocessor.basic_preprocessing(df_temp, data_column)

        papers_temp = papers.copy(deep=True)
        papers_temp = Preprocessor.basic_preprocessing(papers_temp,
                                                       papers_column)

        tfidf_vectorizer = TfidfVectorizer(
            ngram_range=(1, ngrams), max_features=features
        )

        corpus = df_temp["lemma"].tolist() + papers_temp["lemma"].tolist()
        tfidf_vectorizer.fit(corpus)

        data_corpus = df_temp["lemma"].tolist()
        paper_corpus = papers_temp["lemma"].tolist()

        data = tfidf_vectorizer.transform(data_corpus)
        papers = tfidf_vectorizer.transform(paper_corpus)

        return data, papers
